-----------------------------------------------------------------------
-- Upvalued Lua API.
-----------------------------------------------------------------------
local _G = getfenv(0)

-- Functions
local next = _G.next
local pairs = _G.pairs
local time = _G.time
local tonumber = _G.tonumber
local type = _G.type

-- Libraries
local math = _G.math
local table = _G.table

-----------------------------------------------------------------------
-- AddOn namespace.
-----------------------------------------------------------------------
local ADDON_NAME, private = ...

local LibStub = _G.LibStub

local BTC = LibStub("AceAddon-3.0"):NewAddon(ADDON_NAME, "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME)

local LQT = LibStub:GetLibrary("LibQTip-1.0")
local LDBIcon = LibStub("LibDBIcon-1.0")

_G[private.ADDON_NAME_SHORT] = BTC

local addonauthor = "James D. Callahan III"
local builddate = "20121027020307"

------------------------------------------------------------------------------
-- Constants.
------------------------------------------------------------------------------
local ADDON_NAME_SHORT = private.ADDON_NAME_SHORT
local COOLDOWN_SPELL_MAP = private.COOLDOWN_SPELL_MAP
local NUM_TOOLTIP_COLUMNS = 3

local COLOR_TABLE = _G.CUSTOM_CLASS_COLORS or _G.RAID_CLASS_COLORS
local CLASS_COLORS = {}

for class_label, color_data in pairs(COLOR_TABLE) do
	CLASS_COLORS[class_label] = ("%2x%2x%2x"):format(color_data.r * 255, color_data.g * 255, color_data.b * 255)
end

local ICON_MINUS = [[|TInterface\MINIMAP\UI-Minimap-ZoomOutButton-Up:16:16|t]]
local ICON_MINUS_DOWN = [[|TInterface\MINIMAP\UI-Minimap-ZoomOutButton-Down:16:16|t]]

local ICON_PLUS = [[|TInterface\MINIMAP\UI-Minimap-ZoomInButton-Up:16:16|t]]
local ICON_PLUS_DOWN = [[|TInterface\MINIMAP\UI-Minimap-ZoomInButton-Down:16:16|t]]

------------------------------------------------------------------------------
-- Variables.
------------------------------------------------------------------------------
local db
local tooltip

local primary = {}
local secondary = {}


------------------------------------------------------------------------------
-- Helper functions.
------------------------------------------------------------------------------
do
	local CHAT_MESSAGE_FORMAT = "\124TInterface\\Addons\\" .. ADDON_NAME .. "\\icon:" .. "%d" .. "\124t" .. ADDON_NAME_SHORT .. ": " .. "%s"

	function BTC:Print(message)
		if not message then
			return
		end
		_G.DEFAULT_CHAT_FRAME:AddMessage(CHAT_MESSAGE_FORMAT:format(math.ceil(_G.select(2, _G.DEFAULT_CHAT_FRAME:GetFont())), message), 0.6, 1.0, 1.0)
	end
end -- do-block

function BTC:Debug(message)
	if not message or not db.options.debug then
		return
	end
	self:Print("DEBUG: " .. message)
end

-- coloring functions for tooltip text
local function Red(text) return "|cffff0000" .. text .. "|r" end

local function Yellow(text) return "|cffffff00" .. text .. "|r" end

local function Green(text) return "|cff3fbf3f" .. text .. "|r" end

local function CooldownDisplayName(spell_id)
	local display_name = COOLDOWN_SPELL_MAP[spell_id]

	return (type(display_name) == "string") and display_name or _G.select(1, _G.GetSpellInfo(spell_id))
end

local function BTC_prettyTime(seconds)
	local remaining = seconds
	local output = ""

	if remaining >= 86400 then
		output = output .. math.floor(remaining / 86400) .. "d"
		remaining = remaining % 86400
	end

	if remaining >= 3600 then
		output = output .. math.floor(remaining / 3600) .. "h"
		remaining = remaining % 3600
	end

	if remaining >= 60 then
		output = output .. math.floor(remaining / 60) .. "m"
	end
	return output
end

local function ColorizeText(text, r, g, b)
	return ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, text)
end

local function PercentColorText(min, max, text)
	local red_low, green_low, blue_low = 1, 0.10, 0.10
	local red_mid, green_mid, blue_mid = 1, 1, 0
	local red_high, green_high, blue_high = 0.25, 0.75, 0.25
	local percentage = min / max

	if percentage >= 1 then
		return ColorizeText(text, red_high, green_high, blue_high)
	elseif percentage <= 0 then
		return ColorizeText(text, red_low, green_low, blue_low)
	end
	local integral, fractional = math.modf(percentage * 2)

	if integral == 1 then
		red_low, green_low, blue_low, red_mid, green_mid, blue_mid = red_mid, green_mid, blue_mid, red_high, green_high, blue_high
	end
	return ColorizeText(text, red_low + (red_mid - red_low) * fractional, green_low + (green_mid - green_low) * fractional, blue_low + (blue_mid - blue_low) * fractional)
end

do
	local header_list = {}
	local skill_frame_shown

	local function SearchForCooldowns()
		if not skill_frame_shown then
			return
		end
		------------------------------------------------------------------------------
		-- Save the current state of the TradeSkillFrame so it can be restored after we muck with it.
		------------------------------------------------------------------------------
		local have_materials = _G.TradeSkillFrame.filterTbl.hasMaterials
		local have_skillup = _G.TradeSkillFrame.filterTbl.hasSkillUp

		if have_materials then
			_G.TradeSkillFrame.filterTbl.hasMaterials = false
			_G.TradeSkillOnlyShowMakeable(false)
		end

		if have_skillup then
			_G.TradeSkillFrame.filterTbl.hasSkillUp = false
			_G.TradeSkillOnlyShowSkillUps(false)
		end
		_G.SetTradeSkillInvSlotFilter(0, 1, 1)
		_G.TradeSkillUpdateFilterBar()
		_G.TradeSkillFrame_Update()

		------------------------------------------------------------------------------
		-- Expand headers and scan for cooldowns.
		------------------------------------------------------------------------------
		table.wipe(header_list)

		for skill_index = _G.GetNumTradeSkills(), 1, -1 do
			local entry_name, entry_type, _, is_expanded = _G.GetTradeSkillInfo(skill_index)

			if (entry_type == "header" or entry_type == "subheader") and not is_expanded then
				header_list[entry_name] = true
				_G.ExpandTradeSkillSubClass(skill_index)
			end
		end

		for skill_index = 1, _G.GetNumTradeSkills() do
			local entry_name, entry_type = _G.GetTradeSkillInfo(skill_index)

			if entry_type ~= "header" and entry_type ~= "subheader" then
				local spell_id = tonumber(_G.GetTradeSkillRecipeLink(skill_index):match("^|c%x%x%x%x%x%x%x%x|H%w+:(%d+)"))

				if COOLDOWN_SPELL_MAP[spell_id] then
					local cooldown = _G.GetTradeSkillCooldown(skill_index)
					local cooldown_entry = db.cooldowns[spell_id]

					if not cooldown_entry then
						cooldown_entry = {}
						db.cooldowns[spell_id] = cooldown_entry
					end
					local mapped_name = COOLDOWN_SPELL_MAP[spell_id]

					if type(mapped_name) == "string" then
						cooldown_entry.displayName = mapped_name
					else
						cooldown_entry.displayName = entry_name
					end

					if cooldown then
						cooldown_entry.endtime = time() + math.floor(cooldown)
					else
						cooldown_entry.endtime = 0
					end
				else
					-- Just in case we've stored an entry for a spell that was removed from the game.
					db.cooldowns[spell_id] = nil
				end
			end
		end
		------------------------------------------------------------------------------
		-- Restore the state of the things we changed.
		------------------------------------------------------------------------------
		for skill_index = _G.GetNumTradeSkills(), 1, -1 do
			local entry_name, _, _, is_expanded = _G.GetTradeSkillInfo(skill_index)

			if header_list[entry_name] then
				_G.CollapseTradeSkillSubClass(skill_index)
			end
		end
		_G.TradeSkillFrame.filterTbl.hasMaterials = have_materials
		_G.TradeSkillOnlyShowMakeable(have_materials)
		_G.TradeSkillFrame.filterTbl.hasSkillUp = have_skillup
		_G.TradeSkillOnlyShowSkillUps(have_skillup)

		_G.TradeSkillUpdateFilterBar()
		_G.TradeSkillFrame_Update()
	end

	-- I don't actually care about what updated. Just that it did. I only turn on monitoring
	-- for this event when a SPELL_CAST_START has happened for a spell (tradeskill) that has a
	-- cooldown has occurred. And only when it isn't an enchant.
	function BTC:BAG_UPDATE(...)
		BTC:UnregisterEvent("BAG_UPDATE")

		if skill_frame_shown then
			SearchForCooldowns()
		elseif not db.options.hideReminders then
			self:Print(L["CLOSED_TOO_EARLY"])
		end
		private.UpdateDataText()
	end

	function BTC:TRADE_SKILL_SHOW(event_name, ...)
		skill_frame_shown = true

		if _G.IsTradeSkillLinked() or _G.IsTradeSkillGuild() then
			return
		end
		SearchForCooldowns()
		private.UpdateDataText()
	end

	function BTC:TRADE_SKILL_CLOSE(...)
		skill_frame_shown = false
	end
end -- do-block

function BTC:COMBAT_LOG_EVENT_UNFILTERED(_, _, sub_event, _, _, source_name, _, _, _, _, _, _, spell_id)
	if sub_event ~= "SPELL_CAST_START" or source_name ~= private.PLAYER_NAME then
		return
	end
	local now = time()

	local special_case = private.SPECIAL_CASE_COOLDOWNS[spell_id]

	if special_case then
		local entry = db.cooldowns[spell_id]

		if not entry then
			entry = {}
			db.cooldowns[spell_id] = entry
		end
		entry.displayName = special_case.name
		entry.endtime = now + special_case.duration
	elseif COOLDOWN_SPELL_MAP[spell_id] then
		if not db.options.hideReminders then
			self:Print(L["COOLDOWN_DETECTED"])
		end
		-- Right now, all spells with cooldowns place a new item into your bags. Right at the second this
		-- code executes, the player will still be in mid-spell-cast of the spell in question. So in order
		-- to determine when to check for a new cooldown, I'm now going to register for a BAG_UPDATE (the
		-- item hitting my bags), and then scan for new skills.
		self:RegisterEvent("BAG_UPDATE")
	else
		self:Debug("Detected spellcast that wasn't in list. spellid = " .. spell_id)
	end
end

local function UpdateSkills()
	primary[1], primary[2], secondary[1], secondary[2], secondary[3], secondary[4] = _G.GetProfessions()

	for index = 1, 2 do
		if primary[index] then
			local name, _, rank, max_rank = _G.GetProfessionInfo(primary[index])
			db.primary[name] = db.primary[name] or {}
			db.primary[name].currlevel = rank
			db.primary[name].maxlevel = max_rank
		end
	end

	for index = 1, 4 do
		if secondary[index] then
			local name, _, rank, max_rank = _G.GetProfessionInfo(secondary[index])
			db.secondary[name] = db.secondary[name] or {}
			db.secondary[name].currlevel = rank
			db.secondary[name].maxlevel = max_rank
		end
	end
end

local secure_frame = _G.CreateFrame("Button", ADDON_NAME .. "SecureFrame", _G.UIParent, "SecureActionButtonTemplate")
secure_frame:SetAttribute("type", "macro")
secure_frame:SetFrameStrata("TOOLTIP")
secure_frame:Hide()

secure_frame:SetScript("OnLeave", function(self)
	self:SetParent(nil)
	self:ClearAllPoints()
	self:Hide()
end)

local highlight_texture = secure_frame:CreateTexture(nil, "OVERLAY")
highlight_texture:SetTexture([[Interface\QuestFrame\UI-QuestTitleHighlight]])
highlight_texture:SetBlendMode("ADD")
highlight_texture:SetAllPoints(secure_frame)

local function Profession_OnEnter(cell, profession_name)
	secure_frame:SetParent(cell)
	secure_frame:SetPoint("TOPLEFT", cell, "TOPLEFT")
	secure_frame:SetPoint("BOTTOMRIGHT", cell, "BOTTOMRIGHT")
	secure_frame:SetAttribute("macrotext", ("/cast %s\n"):format(profession_name))
	secure_frame:Show()
end

local DrawTooltip
local function Tooltip_OnMouseDown(cell, arg, button)
	local entry_type, toon_name, spell_id = (":"):split(arg)

	spell_id = tonumber(spell_id)

	if _G.IsAltKeyDown() then
		local cooldowns = BTC.db.factionrealm[toon_name].cooldowns

		if entry_type == "name" then
			if button == "LeftButton" then
				for k, v in pairs(cooldowns) do
					cooldowns[k] = nil
				end
			elseif button == "RightButton" then
				for k, v in pairs(cooldowns) do
					v.hide = nil
				end
			end
		elseif entry_type == "spell" then
			if button == "LeftButton" then
				cooldowns[spell_id] = nil
			elseif button == "RightButton" and cooldowns[spell_id] then
				cooldowns[spell_id].hide = true
			end
		end
		DrawTooltip()
	end
	private.UpdateDataText()
end

local function ExpandButton_OnMouseUp(tooltip_cell, character_name)
	local character_entry = BTC.db.factionrealm[character_name]
	character_entry.is_expanded = not character_entry.is_expanded
	DrawTooltip()
end

local function ExpandButton_OnMouseDown(tooltip_cell, is_expanded)
	local line, column = tooltip_cell:GetPosition()
	tooltip:SetCell(line, column, is_expanded and ICON_MINUS_DOWN or ICON_PLUS_DOWN)
end

local GREEN = "|cff3fbf3f"

local function Tooltip_AddProfessionLine(profession_name, data)
	local line_num = tooltip:AddLine()
	tooltip:SetCell(line_num, 1, ("%s%s|r"):format(_G.NORMAL_FONT_COLOR_CODE, profession_name), "LEFT", 2)
	tooltip:SetCell(line_num, 3, ("%s|r/%s%d"):format(PercentColorText(data.currlevel, data.maxlevel, data.currlevel), GREEN, data.maxlevel))
	return line_num
end

local spell_registry = {}

local function CooldownCounterText(target_name)
	local now = time()
	local num_cooldowns = 0
	local num_available = 0

	for character_name, data in pairs(BTC.db.factionrealm) do
		if not target_name or target_name == character_name then
			table.wipe(spell_registry)

			for spell_id, cooldown_data in pairs(data.cooldowns) do
				local display_name = CooldownDisplayName(spell_id)

				if not cooldown_data.hide and not spell_registry[display_name] then
					num_cooldowns = num_cooldowns + 1

					if cooldown_data.endtime == 0 or now > cooldown_data.endtime then
						num_available = num_available + 1
					end
					spell_registry[display_name] = true
				end
			end
		end
	end
	return PercentColorText(num_available, num_cooldowns, _G.PARENS_TEMPLATE:format(num_available .. "/" .. num_cooldowns))
end

function DrawTooltip()
	local now = time()
	local line_num, column

	tooltip:Clear()
	tooltip:SetCellMarginH(0)
	tooltip:SetCellMarginV(1)

	line_num, column = tooltip:AddHeader()
	tooltip:SetCell(line_num, 1, ADDON_NAME, "CENTER", NUM_TOOLTIP_COLUMNS)

	if db.options.showPrimary and next(db.primary) then
		tooltip:AddSeparator()
		line_num = tooltip:AddLine()
		tooltip:SetCell(line_num, 1, _G.PRIMARY_SKILLS, "LEFT", 2)
		tooltip:SetLineColor(line_num, 1, 1, 1, 0.25)
		tooltip:AddSeparator()

		for profession_name, data in pairs(db.primary) do
			line_num = Tooltip_AddProfessionLine(profession_name, data)
			tooltip:SetLineScript(line_num, "OnEnter", Profession_OnEnter, profession_name)
		end
	end

	if db.options.showSecondary and next(db.secondary) then
		tooltip:AddSeparator()
		line_num = tooltip:AddLine()
		tooltip:SetCell(line_num, 1, _G.SECONDARY_SKILLS, "LEFT", 2)
		tooltip:SetLineColor(line_num, 1, 1, 1, 0.25)
		tooltip:AddSeparator()

		for profession_name, data in pairs(db.secondary) do
			line_num = Tooltip_AddProfessionLine(profession_name, data)
			tooltip:SetLineScript(line_num, "OnEnter", Profession_OnEnter, profession_name)
		end
	end
	tooltip:AddSeparator()
	line_num = tooltip:AddLine()
	tooltip:SetCell(line_num, 1, _G.TIME_REMAINING, "LEFT", 2)
	tooltip:SetLineColor(line_num, 1, 1, 1, 0.25)
	tooltip:AddSeparator()

	for character_name, data in pairs(BTC.db.factionrealm) do
		if next(data.cooldowns) then
			local class_color = (type(data.class) == "string") and CLASS_COLORS[data.class] or "cccccc"
			line_num = tooltip:AddLine()
			tooltip:SetCell(line_num, 1, data.is_expanded and ICON_MINUS or ICON_PLUS)
			tooltip:SetCell(line_num, 2, ("|cff%s%s %s|r"):format(class_color, character_name, data.is_expanded and "" or CooldownCounterText(character_name)))
			tooltip:SetCellScript(line_num, 1, "OnMouseUp", ExpandButton_OnMouseUp, character_name)
			tooltip:SetCellScript(line_num, 1, "OnMouseDown", ExpandButton_OnMouseDown, data.is_expanded)
			tooltip:SetCellScript(line_num, 2, "OnMouseDown", Tooltip_OnMouseDown, "name:" .. character_name)
		end
		table.wipe(spell_registry)

		if data.is_expanded then
			for spell_id, cooldown_data in pairs(data.cooldowns) do
				local display_name = CooldownDisplayName(spell_id)

				if not cooldown_data.hide and not spell_registry[display_name] then
					local end_time = cooldown_data.endtime

					if end_time == 0 or now > end_time then
						line_num = tooltip:AddLine(" ", ("%s%s|r"):format(_G.NORMAL_FONT_COLOR_CODE, display_name), Green(_G.READY))
					else
						line_num = tooltip:AddLine(" ", ("%s%s|r"):format(_G.NORMAL_FONT_COLOR_CODE, display_name), PercentColorText(end_time - now, end_time, BTC_prettyTime(end_time - now)))
					end
					tooltip:SetLineScript(line_num, "OnMouseDown", Tooltip_OnMouseDown, "spell:" .. character_name .. ":" .. spell_id)
					spell_registry[display_name] = true
				end
			end
		end
	end
	tooltip:AddSeparator()

	line_num = tooltip:AddLine()
	tooltip:SetCell(line_num, 2, L["TOOLTIP_HINT_1"], "LEFT", NUM_TOOLTIP_COLUMNS - 1)

	line_num = tooltip:AddLine(" ")
	tooltip:SetCell(line_num, 2, L["TOOLTIP_HINT_2"], "LEFT", NUM_TOOLTIP_COLUMNS - 1)

	line_num = tooltip:AddLine(" ")
	tooltip:SetCell(line_num, 2, L["TOOLTIP_HINT_3"], "LEFT", NUM_TOOLTIP_COLUMNS - 1)
end

local LDB = LibStub("LibDataBroker-1.1", true)
local launcher = LDB:NewDataObject("BTC", {
	type = "data source",
	text = "",
	label = "Broker_TradeCooldowns",
	icon = "Interface\\Addons\\Broker_TradeCooldowns\\icon",
	OnClick = function(clickedframe, button)
		_G.InterfaceOptionsFrame_OpenToCategory(BTC.optionsFrame["Display"])
		private.UpdateDataText()
	end,
	OnEnter = function(self)
		tooltip = LQT:Acquire(ADDON_NAME .. "Tooltip", NUM_TOOLTIP_COLUMNS, "LEFT", "LEFT", "RIGHT")
		tooltip:SetScale(db.options.tipscale)
		DrawTooltip()
		tooltip:SetAutoHideDelay(.1, self)
		tooltip:EnableMouse()
		tooltip:SmartAnchorTo(self)
		tooltip:Show()
	end,
	OnLeave = function()
	-- this should get handled by the autohide delay stuff now
	end,
})

-- options stuff
local displayOptions = nil
local function giveDisplay()
	if (not displayOptions) then
		displayOptions = {
			order = 4,
			type = "group",
			name = L["Display Options"],
			desc = L["DISPLAY_OPTIONS_DESC"],
			args = {
				header1 = {
					order = 41,
					type = "header",
					name = L["Display Options"],
				},
				desc = {
					order = 42,
					type = "description",
					name = L["DISPLAY_OPTIONS_DESC"],
				},
				tooltipscale = {
					order = 43,
					type = "range",
					name = L["Tooltip Scale"],
					desc = L["TOOLTIP_SCALE_DESC"],
					min = .7,
					max = 1.3,
					step = .05,
					bigStep = .05,
					get = function() return db.options.tipscale end,
					set = function(info, v)
						db.options.tipscale = v
						tooltip:SetScale(v)
					end,
				},
				--[===[@debug@
				debug = {
					order = 44,
					type = "toggle",
					name = L["Debug"],
					desc = L["DEBUG_DESC"],
					get = function() return db.options.debug end,
					set = function() db.options.debug = not db.options.debug end,
				},
				--@end-debug@]===]
				showprimary = {
					order = 45,
					type = "toggle",
					name = L["Show Primary"],
					desc = L["SHOW_PRIMARY_DESC"],
					get = function() return db.options.showPrimary end,
					set = function() db.options.showPrimary = not db.options.showPrimary end,
				},
				showsecondary = {
					order = 46,
					type = "toggle",
					name = L["Show Secondary"],
					desc = L["SHOW_SECONDARY_DESC"],
					get = function() return db.options.showSecondary end,
					set = function() db.options.showSecondary = not db.options.showSecondary end,
				},
				hidereminders = {
					order = 47,
					type = "toggle",
					name = L["Hide Reminders"],
					desc = L["HIDE_REMINDERS_DESC"],
					get = function() return db.options.hideReminders end,
					set = function() db.options.hideReminders = not db.options.hideReminders end,
				},
				header2 = {
					order = 200,
					type = "header",
					name = L["Minimap Icon Options"],
				},
				desc2 = {
					order = 201,
					type = "description",
					name = L["MINIMAP_OPTIONS_DESC"],
				},
				minimap_icon = {
					order = 202,
					type = "toggle",
					width = "full",
					name = L["Show Minimap Icon"],
					desc = L["Draws the icon on the minimap."],
					get = function()
						return not db.options.minimap_icon.hide
					end,
					set = function(info, value)
						db.options.minimap_icon.hide = not value
						LDBIcon[value and "Show" or "Hide"](LDBIcon, ADDON_NAME_SHORT)
					end,
				},
				header3 = {
					order = 301,
					type = "header",
					name = L["Help and Usage Notes"],
				},
				help1 = {
					order = 302,
					type = "description",
					name = "- " .. L["HELP_STRING_0"],
				},
				help2 = {
					order = 303,
					type = "description",
					name = "- " .. L["HELP_STRING_1"],
				},
				help3 = {
					order = 304,
					type = "description",
					name = "- " .. L["HELP_STRING_2"],
				},
				help4 = {
					order = 305,
					type = "description",
					name = "- " .. L["HELP_STRING_3"],
				},
				help5 = {
					order = 306,
					type = "description",
					name = "- " .. L["HELP_STRING_4"],
				},
			},
		}
	end
	return displayOptions
end

local textOptions = nil
local function giveText()
	if (not textOptions) then
		textOptions = {
			order = 4,
			type = "group",
			name = L["Text Options"],
			desc = L["TEXT_OPTIONS_DESC"],
			args = {
				header1 = {
					order = 41,
					type = "header",
					name = L["Text Options"],
				},
				desc = {
					order = 42,
					type = "description",
					name = L["TEXT_OPTIONS_DESC"],
				},
				showtext = {
					order = 50,
					type = "toggle",
					name = L["Show Text"],
					desc = L["SHOW_TEXT_DESC"],
					get = function() return db.options.showText end,
					set = function()
						db.options.showText = not db.options.showText
						private.UpdateDataText()
					end,
				},
				showcounters = {
					order = 51,
					type = "toggle",
					name = L["Show Counters"],
					desc = L["SHOW_COUNTERS_DESC"],
					get = function() return db.options.showCounters end,
					set = function()
						db.options.showCounters = not db.options.showCounters
						private.UpdateDataText()
					end,
					disabled = function()
						if (db.options.showText) then
							return false
						else
							return true
						end
					end,
				},
				textcooldown = {
					order = 52,
					type = "select",
					name = L["Select Cooldown to Display"],
					desc = L["SHOW_TEXTCD_DESC"],
					get = function() return db.options.textCooldown end,
					set = function(info, value)
						db.options.textCooldown = value
						private.UpdateDataText()
					end,
					values = function()
						local t = {}

						for character_name, data in pairs(BTC.db.factionrealm) do
							table.wipe(spell_registry)

							for spell_id, cooldown_data in pairs(data.cooldowns) do
								local display_name = CooldownDisplayName(spell_id)

								if not cooldown_data.hide and not spell_registry[display_name] then
									t[character_name .. ":" .. spell_id] = _G.SUBTITLE_FORMAT:format(character_name, display_name)
									spell_registry[display_name] = true
								end
							end
						end
						return t
					end,
					disabled = function()
						return not db.options.showText or db.options.showCounters
					end,
				},
			},
		}
	end
	return textOptions
end

local modularOptions = {}
local options = nil
local function fullOptions()
	if (not options) then
		options = {
			type = "group",
			name = ADDON_NAME_SHORT,
			args = {
				general = {
					order = 1,
					type = "group",
					name = L["General Information"],
					desc = L["GENERAL_INFO_DESC"],
					args = {
						header1 = {
							order = 11,
							type = "header",
							name = L["General Information"],
						},
						desc = {
							order = 12,
							type = "description",
							name = L["GENERAL_INFO_DESC"],
						},
						author = {
							order = 13,
							type = "description",
							name = L["Author : "] .. addonauthor .. "\n",
						},
						version = {
							order = 14,
							type = "description",
							name = L["Version : "] .. private.ADDON_VERSION .. "\n",
						},
						builddate = {
							order = 15,
							type = "description",
							name = L["Build Date : "] .. builddate .. "\n",
						},
						translators = {
							order = 16,
							type = "description",
							name = string.format(L["Helpful Translators (thank you) : %s"],
								"whynona2(frFR), SunTsu(deDE), Dessa(deDE), " ..
									"amunra(koKR), xska(ruRU), <CWDG>tnt2ray(zhCN), " ..
									"NightOw1(zhTW), andy52005(zhTW)\n")
						},
					},
				},
			},
		}
	end

	for k, v in pairs(modularOptions) do
		options.args[k] = (type(v) == "function") and v() or v
	end

	return options
end

local AceConfig = LibStub("AceConfig-3.0")
local AceConfigReg = LibStub("AceConfigRegistry-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")

function BTC:SetupOptions()
	AceConfigReg:RegisterOptionsTable(ADDON_NAME_SHORT, fullOptions)
	self.optionsFrame = AceConfigDialog:AddToBlizOptions(ADDON_NAME_SHORT, nil, nil, "general")

	-- Fill up our modular options...
	self:RegisterModuleOptions("Display", giveDisplay(), L["Display Options"])
	self:RegisterModuleOptions("Text", giveText(), L["Text Options"])
end

function BTC:RegisterModuleOptions(name, optionsTable, displayName)
	modularOptions[name] = optionsTable
	self.optionsFrame[name] = AceConfigDialog:AddToBlizOptions(ADDON_NAME_SHORT, displayName, ADDON_NAME_SHORT, name)
end

function private.UpdateDataText()
	if not db.options.showText then
		launcher.text = " "
		return
	end

	if db.options.showCounters then
		launcher.text = CooldownCounterText()
		return
	end

	if db.options.textCooldown == 0 then
		launcher.text = " "
		return
	end
	local target_name, target_spell_id = (":"):split(db.options.textCooldown)
	local now = time()

	if not target_name or target_name == "" or not target_spell_id then
		launcher.text = " "
		return
	end

	for character_name, data in pairs(BTC.db.factionrealm) do
		if character_name == target_name then
			for spell_id, cooldown_data in pairs(data.cooldowns) do
				target_spell_id = target_spell_id + 0

				if spell_id == target_spell_id then
					local end_time = cooldown_data.endtime
					local time_text

					if end_time == 0 or now > end_time then
						time_text = Green(_G.READY)
					else
						time_text = PercentColorText(end_time - now, end_time, BTC_prettyTime(end_time - now))
					end
					launcher.text = _G.SUBTITLE_FORMAT:format(Yellow(COOLDOWN_SPELL_MAP[spell_id]), time_text)
					return
				end
			end
		end
	end
end

-- TODO: Get rid of this. Seriously.
local BTC_ScanFrame

function private.ScanSkills()
	-- on first load, there's no real good way to check if the player's skills
	-- are available for scanning yet. So just scan every 5 seconds until they are
	if (not BTC_ScanFrame) then
		BTC_ScanFrame = _G.CreateFrame("Frame")
	end

	local totalElapsed = 0
	local checkInterval = 60
	local gotFirst = false

	BTC_ScanFrame:SetScript("OnUpdate", function(self, elapsed)
		totalElapsed = totalElapsed + elapsed
		if (totalElapsed < checkInterval) then
			return
		end
		totalElapsed = 0
		UpdateSkills()
		private.UpdateDataText()
		return
	end)
end

function BTC:OnInitialize()
end

function BTC:OnEnable()
	self.db = LibStub("AceDB-3.0"):New("BTC_DB", nil, "Default")
	if not self.db.factionrealm then
		self.db.factionrealm = {}
	end

	-- do we have any data yet for this toon? If not, init its table
	if not (self.db.factionrealm[private.PLAYER_NAME]) then
		self.db.factionrealm[private.PLAYER_NAME] = {
			primary = {},
			secondary = {},
			class = {},
			languages = {},
			armor = {},
			cooldowns = {},
			options = {
				dbVersion = "1.8",
				debug = false,
				showPrimary = true,
				showSecondary = true,
				showText = false,
				showCounters = false,
				hideReminders = false,
				textCooldown = 0,
				tipscale = 1,
				minimap_icon = {
					hide = true,
				},
			}
		}
	end

	db = self.db.factionrealm[private.PLAYER_NAME]
	db.class = _G.select(2, _G.UnitClass("player"))

	BTC:SetupOptions()

	-- Check current weapons/languages
	private.ScanSkills()
	private.UpdateDataText()

	-- Event listeners
	BTC:RegisterEvent("TRADE_SKILL_SHOW")
	BTC:RegisterEvent("TRADE_SKILL_CLOSE")
	BTC:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	BTC:RegisterEvent("CHAT_MSG_SKILL", UpdateSkills)

	-- libdbicon stuff
	if LDBIcon then
		LDBIcon:Register(ADDON_NAME_SHORT, launcher, db.options.minimap_icon)
	end
end
