-----------------------------------------------------------------------
-- Upvalued Lua API.
-----------------------------------------------------------------------
local _G = getfenv(0)

-- Functions
local next = _G.next
local pairs = _G.pairs
local time = _G.time
local tonumber = _G.tonumber
local type = _G.type

-- Libraries
local math = _G.math
local table = _G.table

-----------------------------------------------------------------------
-- AddOn namespace.
-----------------------------------------------------------------------
local ADDON_NAME, private = ...

local LibStub = _G.LibStub
local L = LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME)

------------------------------------------------------------------------------
-- Constants.
------------------------------------------------------------------------------
private.ADDON_NAME_SHORT = "BTC"
private.PLAYER_NAME = _G.UnitName("player")

do
	local debug_version, alpha_version

	--[===[@debug@
	debug_version = true
	--@end-debug@]===]

	--@alpha@
	alpha_version = true
	--@end-alpha@

	local addon_version = _G.GetAddOnMetadata(ADDON_NAME, "Version")
	private.ADDON_VERSION = debug_version and "Devel" or (alpha_version and addon_version .. "-Alpha") or addon_version
end -- do-block

private.SPECIAL_CASE_COOLDOWNS = {
	-- Snowmaster 9000
	[21935] = {
		name = L["Snowmaster 9000"],
		duration = 72000,
	},
	-- Create Elune Stone
	[26265] = {
		name = L["Elune's Lantern"],
		duration = 72000,
	},
}

private.COOLDOWN_SPELL_MAP = {
	-- Initialize my cooldowns list
	-- Alchemy
	-- Northrend Alchemy Research
	[60893] = true, -- Northrend Alchemy Research
	-- Metals (shared cooldown)
	[11479] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Iron to Gold
	[11480] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Mithril to Truesilver
	-- rm'd in 3.3.0
	-- [60350] = _G.STRING_SCHOOL_ELEMENTAL,		-- Transmute: Titanium
	-- Elements (shared cooldown)
	[17559] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Air to Fire
	[17566] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Earth to Life
	[17561] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Earth to Water
	--[[ No cooldown on Elemental Fire
			 [25146] = _G.STRING_SCHOOL_ELEMENTAL,
			 ]] --
	[17560] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Fire to Earth
	[17565] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Life to Earth
	[17563] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Undeath to Water
	[17562] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Water to Air
	[17564] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Water to Undeath
	-- Primals
	[28566] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Air to Fire
	[28585] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Earth to Life
	[28567] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Earth to Water
	[28568] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Fire to Earth
	[28583] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Fire to Mana
	[28584] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Life to Earth
	[28582] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Mana to Fire
	[28580] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Shadow to Water
	[28569] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Water to Air
	[28581] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Primal Water to Shadow
	-- Eternals
	[53777] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Air to Earth
	[52776] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Air to Water
	[53781] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Earth to Air
	[53782] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Earth to Shadow
	[53775] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Fire to Life
	[53774] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Fire to Water
	[53773] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Life to Fire
	[53771] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Life to Shadow
	[54020] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Might
	[53779] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Shadow to Earth
	[52780] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Shadow to Life
	[53783] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Water to Air
	[53784] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Eternal Water to Fire
	-- Epic gems
	[66658] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute : Ametrine
	[66659] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute : Cardinal Ruby
	[66660] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute : King's Amber
	[66662] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute : Dreadstone
	[66663] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute : Majestic Zircon
	[66664] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute : Eye of Zul
	-- Diamonds (shared cooldown)
	--[[ It appears these were removed as well
			 [32765] = L["Diamonds"],		-- Transmute: Earthstorm Diamond
			 [32766] = L["Diamonds"],		-- Transmute: Skyfire Diamond
			 [57425] = L["Diamonds"],		-- Transmute: Skyflare Diamond
			 [57427] = L["Diamonds"],		-- Transmute: Earthsiege Diamond
			 ]] --
	-- New Cata Elemental transmutes
	[78866] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Living Elements
	[80243] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Truegold
	[80244] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Pyrium Bar
	-- MoP
	[114780] = _G.STRING_SCHOOL_ELEMENTAL, -- Transmute: Living Steel

	-- New Cata Gems
	--[[ Removed somewhere around 4.1
			 [80245] = L["Cataclysm Blue Gems"],		-- Transmute: Inferno Ruby
			 [80246] = L["Cataclysm Blue Gems"],		-- Transmute: Ocean Sapphire
			 [80247] = L["Cataclysm Blue Gems"],		-- Transmute: Amberjewel
			 [80248] = L["Cataclysm Blue Gems"],		-- Transmute: Demonseye
			 [80250] = L["Cataclysm Blue Gems"],		-- Transmute: Ember Topaz
			 [80251] = L["Cataclysm Blue Gems"],		-- Transmute: Dream Emerald
			 ]] --

	-- Enchanting
	[28027] = true, -- Prismatic Sphere
	[28028] = true, -- Void Sphere
	[116499] = true, -- Sha Crystal

	-- Jewelcrafting
	[47280] = true, -- Brilliant Glass
	[62242] = true, -- Icy Prism
	[73478] = true, -- Fire Prism

	-- MoP Jewelcrafting (shared cooldown)
	[131691] = L["Facets of Research"], -- Imperial Amethyst
	[131686] = L["Facets of Research"], -- Primordial Ruby
	[131593] = L["Facets of Research"], -- River's Heart
	[131695] = L["Facets of Research"], -- Sun's Radiance
	[131690] = L["Facets of Research"], -- Vermilion Onyx
	[131688] = L["Facets of Research"], -- Wild Jade

	-- Tailoring
	-- Wrath of the Lich King
	[56005] = true, -- Glacial Bag
	-- Cataclysm
	[75141] = true, -- Dream of Skywall
	[75145] = true, -- Dream of Ragnaros
	[75144] = true, -- Dream of Hyjal
	[75142] = true, -- Dream of Deepholm
	[75146] = true, -- Dream of Azshara
	-- Mists of Pandaria
	[125557] = true, -- Imperial Silk
	-- Inscription
	[61288] = true, -- Minor Inscription Research
	[61177] = true, -- Northrend Inscription Research
	[86654] = true, -- Horde Forged Documents
	[89244] = true, -- Alliance Forged Documents
	[112996] = true, -- Scroll of Wisdom
}
